@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 25336 $
+ * $Revision: 27987 $
  * $Author: rjongbloed $
- * $Date: 2011-03-16 03:51:39 -0500 (Wed, 16 Mar 2011) $
+ * $Date: 2012-07-10 06:06:01 -0500 (Tue, 10 Jul 2012) $
  */
 
 #include <ptlib.h>
@@ -60,7 +60,6 @@ OpalCall::OpalCall(OpalManager & mgr)
   , isEstablished(PFalse)
   , isClearing(PFalse)
   , callEndReason(OpalConnection::NumCallEndReasons)
-  , endCallSyncPoint(NULL)
 #if OPAL_HAS_MIXER
   , m_recordManager(NULL)
 #endif
@@ -103,7 +102,8 @@ void OpalCall::OnEstablishedCall()
 void OpalCall::SetCallEndReason(OpalConnection::CallEndReason reason)
 {
   // Only set reason if not already set to something
-  if (callEndReason == OpalConnection::NumCallEndReasons)
+  if (callEndReason == OpalConnection::NumCallEndReasons &&
+     (reason != OpalConnection::EndedByCallForwarded || connectionsActive.GetSize() <= 1))
     callEndReason = reason;
 }
 
@@ -112,49 +112,74 @@ void OpalCall::Clear(OpalConnection::CallEndReason reason, PSyncPoint * sync)
 {
   PTRACE(3, "Call\tClearing " << (sync != NULL ? "(sync) " : "") << *this << " reason=" << reason);
 
-  if (!LockReadWrite())
-    return;
+  {
+    PSafeLockReadWrite lock(*this);
+    if (!lock.IsLocked() || isClearing) {
+      if (sync != NULL)
+        sync->Signal();
+      return;
+    }
 
-  isClearing = PTrue;
+    isClearing = true;
 
-  SetCallEndReason(reason);
+    SetCallEndReason(reason);
 
-  if (sync != NULL && !connectionsActive.IsEmpty()) {
-    // only set the sync point if it is NULL
-    if (endCallSyncPoint == NULL)
-      endCallSyncPoint = sync;
-    else {
-      PAssertAlways("Can only have one thread doing ClearCallSynchronous");
+    if (sync != NULL)
+      m_endCallSyncPoint.push_back(sync);
+
+    switch (connectionsActive.GetSize()) {
+      case 0 :
+        break;
+
+      case 1 :
+        {
+          PSafePtr<OpalConnection> connection = connectionsActive.GetAt(0, PSafeReference);
+          if (connection != NULL)
+            connection->Release(reason);
+        }
+        break;
+
+      default :
+        // Release all but A-Party, it gets done in the B-Party OnReleased thread.
+        for (PINDEX i = 1; i < connectionsActive.GetSize(); ++i) {
+          PSafePtr<OpalConnection> connection = connectionsActive.GetAt(i, PSafeReference);
+          if (connection != NULL)
+            connection->Release(reason);
+        }
     }
   }
 
-  UnlockReadWrite();
-
+  // Outside of lock
   InternalOnClear();
+}
 
-  PSafePtr<OpalConnection> connection;
-  while (EnumerateConnections(connection, PSafeReadWrite))
-    connection->Release(reason);
+
+void OpalCall::InternalOnClear()
+{
+  if (connectionsActive.IsEmpty() && manager.activeCalls.Contains(GetToken())) {
+    OnCleared();
+
+#if OPAL_HAS_MIXER
+    StopRecording();
+#endif
+
+    if (LockReadWrite()) {
+      while (!m_endCallSyncPoint.empty()) {
+        PTRACE(5, "Call\tSignalling end call.");
+        m_endCallSyncPoint.front()->Signal();
+        m_endCallSyncPoint.pop_front();
+      }
+      UnlockReadWrite();
+    }
+
+    manager.activeCalls.RemoveAt(GetToken());
+  }
 }
 
 
 void OpalCall::OnCleared()
 {
   manager.OnClearedCall(*this);
-
-#if OPAL_HAS_MIXER
-  StopRecording();
-#endif
-
-  if (!LockReadWrite())
-    return;
-
-  if (endCallSyncPoint != NULL) {
-    endCallSyncPoint->Signal();
-    endCallSyncPoint = NULL;
-  }
-
-  UnlockReadWrite();
 }
 
 
@@ -347,7 +372,7 @@ bool OpalCall::Transfer(const PString & newAddress, OpalConnection * connection)
 
   if (connection == NULL) {
     for (PSafePtr<OpalConnection> conn = GetConnection(0); conn != NULL; ++conn) {
-      if (prefix == conn->GetPrefixName())
+      if (prefix == conn->GetPrefixName() && !conn->IsReleased())
         return conn->TransferConnection(newAddress);
     }
 
@@ -355,6 +380,11 @@ bool OpalCall::Transfer(const PString & newAddress, OpalConnection * connection)
     return false;
   }
 
+  if (connection->IsReleased()) {
+    PTRACE(2, "Call\tCannot transfer to released connection " << *connection);
+    return false;
+  }
+
   if (prefix == "*")
     return connection->TransferConnection(connection->GetPrefixName() + newAddress.Mid(1));
 
@@ -420,8 +450,8 @@ OpalMediaFormatList OpalCall::GetMediaFormats(const OpalConnection & connection)
 
   connection.AdjustMediaFormats(true, NULL, commonFormats);
 
-  PTRACE(4, "Call\tGetMediaFormats for " << connection << '\n'
-         << setfill('\n') << commonFormats << setfill(' '));
+  PTRACE(4, "Call\tGetMediaFormats for " << connection << "\n    "
+         << setfill(',') << commonFormats << setfill(' '));
 
   return commonFormats;
 }
@@ -590,11 +620,12 @@ PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection,
       sourceStream = connection.OpenMediaStream(sourceFormat, sessionID, true);
       if (sourceStream == NULL)
         return false;
+      // If re-opening, need to disconnect from old patch. If new then this sets NULL to NULL.
+      sourceStream->SetPatch(NULL);
     }
 
     sinkStream = otherConnection->OpenMediaStream(sinkFormat, sessionID, false);
     if (sinkStream != NULL) {
-      startedOne = true;
       if (!sourceStream.SetSafetyMode(PSafeReadOnly))
         return false;
 
@@ -606,8 +637,12 @@ PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection,
           return false;
       }
 
-      patch->AddSink(sinkStream);
       sourceStream.SetSafetyMode(PSafeReference);
+
+      if (patch->AddSink(sinkStream))
+        startedOne = true;
+      else
+        sinkStream->Close();
     }
   }
 
@@ -729,22 +764,13 @@ void OpalCall::OnReleased(OpalConnection & connection)
   {
     PSafePtr<OpalConnection> last = connectionsActive.GetAt(0, PSafeReference);
     if (last != NULL)
-      last->Release(connection.GetCallEndReason());
+      last->Release(connection.GetCallEndReason(), true);
   }
 
   InternalOnClear();
 }
 
 
-void OpalCall::InternalOnClear()
-{
-  if (connectionsActive.IsEmpty() && manager.activeCalls.Contains(GetToken())) {
-    OnCleared();
-    manager.activeCalls.RemoveAt(GetToken());
-  }
-}
-
-
 void OpalCall::OnHold(OpalConnection & /*connection*/,
                       bool /*fromRemote*/, 
                       bool /*onHold*/)
